/**
  * date-format-plugin v0.1.7
  *
  * @link https://vue-formily.netlify.app/plugins/date-format
  * @source https://github.com/vue-formily/date-format-plugin
  * (c) 2022 An Ha
  * @license MIT
  */
(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
  typeof define === 'function' && define.amd ? define(factory) :
  (global = typeof globalThis !== 'undefined' ? globalThis : global || self, global.DateFormatPlugin = factory());
})(this, (function () { 'use strict';

  const _toString = Object.prototype.toString;
  function isNullOrUndefined(value) {
    return value === null || value === undefined;
  }
  /**
   * Strict object type check. Only returns true
   * for plain JavaScript objects.
   */

  function isPlainObject(value) {
    return _toString.call(value) === '[object Object]';
  }
  function isNumber(value) {
    return typeof value === 'number';
  }
  function isString(value) {
    return typeof value === 'string';
  }

  const rkey = /([^[\]]+)/g;
  /**
   * Find the first found value in objects or arrays by its path
   */

  function get(path, ...args) {
    const length1 = args.length;
    const paths = isString(path) ? path.split('.') : path;
    const length2 = paths.length;

    for (let i = 0; i < length1; i++) {
      let value = args[i];
      let j = -1;

      while (value && ++j < length2) {
        const fullName = paths[j];
        const m = fullName.match(rkey);

        if (m) {
          const length3 = m.length;

          for (let k = 0; k < length3; k++) {
            const name = m[k];

            if ((Array.isArray(value) || isPlainObject(value)) && name in value) {
              value = value[name];

              if (j === length2 - 1 && k === length3 - 1) {
                return value;
              }
            }
          }
        }
      }
    }

    return undefined;
  }

  const ladder = [0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334];
  /**
   * Hold monday column values then generate the rest days on this value
   * M T W T F S S
   * 1 2 3 4 5 6 7
   * 7 1 2 3 4 5 6
   * 6 7 1 2 3 4 5
   * 5 6 7 1 2 3 4
   * 4 5 6 7 1 2 3
   * 3 4 5 6 7 1 2
   * 2 3 4 5 6 7 1
   */
  const weeksMap = [1, 7, 6, 5, 4, 3, 2];
  const rsep = /^[, \u200e]+/;
  const rtimezone = /([+-])(\d{1,2})(?::(\d{1,2}))?/;
  const MINUTE = 60000;
  const HOUR = 60 * MINUTE;
  function isLeapYear(year) {
      return year % 4 === 0 && (year % 100 !== 0 || year % 400 === 0);
  }
  /**
   * Ref: https://en.wikipedia.org/wiki/ISO_week_date
   * The number of weeks in a given year is equal to the corresponding week number of 28 December, because it is the only date that is always in the last week of the year since it is a week before 4 January which is always in the first week of the following year.
   *
   * Using only the ordinal year number y, the number of weeks in that year can be determined:[1]
   *
   * p(y) = (y + (y /4) - (y / 100) + (y / 400)) % 7
   * weeks(y) = 52 + (p(y) === 4 || p(y - 1) === 3 ? 1 : 0)
   */
  function wiy(year) {
      return (year + Math.floor(year / 4) - Math.floor(year / 100) + Math.floor(year / 400)) % 7;
  }
  function computeWeeksInYear(year) {
      const wiy1 = wiy(year);
      const wiy2 = wiy(year - 1);
      return 52 + (wiy1 === 4 || wiy2 === 3 ? 1 : 0);
  }
  function computeOrdinal({ year, month, day }) {
      return day + ladder[month - 1] + (isLeapYear(year) && month > 2 ? 1 : 0);
  }
  function gregorianToWeek(cal) {
      const { year, month, day, minimalDaysInFirstWeek } = cal;
      const ordinal = computeOrdinal({ year, month, day });
      /**
       * Ref: https://en.wikipedia.org/wiki/ISO_week_date
       * The week number (WW or woy for week of year) of any date can be calculated, given its ordinal date (i.e. day of the year, doy or DDD, 1–365 or 366) and its day of the week (D or dow, 1–7).
       * woy = (10 + doy − dow) div 7
       * where
       * doy = 1 → 365/366, dow = 1 → 7 og div means integer division (i.e. the remainder after a division is discarded).
       */
      let weekInYear = Math.floor((14 - minimalDaysInFirstWeek + ordinal - cal.getDayOfWeek()) / 7);
      let weekYear;
      if (weekInYear < 1) {
          weekYear = year - 1;
          weekInYear = computeWeeksInYear(weekYear);
      }
      else if (weekInYear > computeWeeksInYear(year)) {
          weekYear = year + 1;
          weekInYear = 1;
      }
      else {
          weekYear = year;
      }
      return { weekYear, weekInYear };
  }
  function dtfFormat({ date, timeZoneName, locale = 'en-US', timeZone } = {}) {
      return new Intl.DateTimeFormat(locale, {
          timeZoneName,
          timeZone
      }).format(date);
  }
  function extractTimeZoneOffset(timeZone) {
      const match = timeZone.match(rtimezone);
      if (match) {
          const [, sign, hours, minutes] = match;
          return (+hours * HOUR + (minutes ? +minutes * MINUTE : 0)) * (sign === '+' ? 1 : -1);
      }
      return 0;
  }
  function parseTime(offset) {
      const seconds = Math.trunc(Math.abs(offset / 1000));
      const hours = Math.trunc(Math.abs(seconds / 3600));
      const minutes = Math.trunc(Math.abs(seconds % 60));
      const sign = offset >= 0 ? '+' : '-';
      return {
          hours,
          minutes,
          seconds,
          sign
      };
  }
  class Calendar {
      constructor(timestamp, options = {}) {
          this.timestamp = (timestamp instanceof Date ? timestamp.getTime() : timestamp) || Date.now();
          this._d = {};
          const { locale = 'en-US', minimalDaysInFirstWeek = 4, firstDayOfWeek = Calendar.MONDAY } = options;
          this.firstDayOfWeek = +firstDayOfWeek;
          this.minimalDaysInFirstWeek = +minimalDaysInFirstWeek;
          this.locale = locale;
          this.setTimeZone(options.timeZone);
      }
      get offset() {
          return this._d.offset;
      }
      get timeZone() {
          return this._d.timeZone;
      }
      setTimeZone(val) {
          let os;
          let tz;
          if (isNumber(val)) {
              os = val;
              tz = 'UTC';
          }
          else {
              if (isNullOrUndefined(val)) {
                  tz = new Intl.DateTimeFormat().resolvedOptions().timeZone;
              }
              else if (this.isValidTimeZone(val)) {
                  tz = val;
              }
              else {
                  throw new Error('invalid timezone format');
              }
              const formatted = dtfFormat({ date: this.instance, timeZoneName: 'short', timeZone: val });
              os = extractTimeZoneOffset(formatted);
          }
          const _d = this._d;
          _d.offset = os;
          _d.timeZone = tz;
          _d.instance = new Date(this.timestamp + _d.offset);
      }
      get instance() {
          return this._d.instance;
      }
      get year() {
          return this.instance.getUTCFullYear();
      }
      get month() {
          return this.instance.getUTCMonth() + 1;
      }
      get day() {
          return this.instance.getUTCDate();
      }
      get hour() {
          return this.instance.getUTCHours();
      }
      get minute() {
          return this.instance.getUTCMinutes();
      }
      get second() {
          return this.instance.getUTCSeconds();
      }
      get millisecond() {
          return this.instance.getUTCMilliseconds();
      }
      isValidTimeZone(timeZone) {
          try {
              dtfFormat({ timeZone });
              return true;
          }
          catch (e) {
              return false;
          }
      }
      getWeekInYear() {
          return gregorianToWeek(this).weekInYear;
      }
      getWeekYear() {
          return gregorianToWeek(this).weekYear;
      }
      getDayInYear() {
          let diy = ladder[this.month - 1] + this.day;
          if (this.month > 2 && isLeapYear(this.year)) {
              diy += 1;
          }
          return diy;
      }
      getDayOfWeek() {
          const utcDay = this.instance.getUTCDay();
          let i = weeksMap[this.firstDayOfWeek];
          const map = weeksMap.map(() => {
              if (i > 7) {
                  i = 1;
              }
              return i++;
          });
          return map[utcDay === 0 ? 6 : utcDay - 1];
      }
      getTimeZoneName({ format } = {}) {
          const { instance: date, timeZone, locale } = this;
          const short = dtfFormat({ date, timeZone });
          const long = dtfFormat({ date, timeZone, locale, timeZoneName: format });
          const diffed = long.substring(short.length);
          const trimmed = diffed.replace(rsep, '');
          return trimmed;
      }
  }
  Calendar.MONDAY = 0;
  Calendar.TUESDAY = 1;
  Calendar.WEDNESDAY = 2;
  Calendar.THURSDAY = 3;
  Calendar.FRIDAY = 4;
  Calendar.SATURDAY = 5;
  Calendar.SUNDAY = 6;

  // - (\w)\1* matches any sequences of the same letter
  // - '' matches two quote characters in a row
  // - '(''|[^'])+('|$) matches anything surrounded by two quote characters ('),
  //   except a single quote symbol, which ends the sequence.
  const formattingTokensRegExp = /(\w)\1*|''|'(''|[^'])+('|$)|./g;
  const escapedStringRegExp = /^'([^]*?)'?$/;
  const doubleQuoteRegExp = /''/g;
  function zeroPad(input, targetLength) {
      const num = +input;
      const sign = num < 0 ? '-' : '';
      const length = ('' + Math.abs(num)).length;
      return targetLength > length
          ? `${sign}${Array(targetLength)
            .concat([Math.abs(num)])
            .join('0')
            .slice(-targetLength)}`
          : '' + input;
  }
  function cleanEscapedString(input) {
      const match = input.match(escapedStringRegExp);
      return match ? match[1].replace(doubleQuoteRegExp, "'") : input;
  }
  function formatYear(input, token) {
      const length = token.length;
      return zeroPad(length === 2 ? ('' + input).slice(-2) : input, length) + '';
  }
  const _lengthNames = ['narrow', 'short', 'long'];
  function getLengthName(length) {
      return _lengthNames[Math.min(2, length > 0 ? length - 1 : 0)];
  }
  function weekInMonth(cal) {
      const dow = cal.getDayOfWeek();
      const remain = cal.day - dow;
      return Math.floor(remain > 0 ? 1 + remain / 7 + (remain % 7 !== 0 ? 1 : 0) : 1);
  }
  function localize(path, { locales = [], locale } = {}) {
      const _locale = locales.find(loc => loc.code === locale) || {};
      const value = get(path, _locale.localize);
      return value !== undefined ? '' + value : path;
  }
  const formatters = {
      // Era designator, e.g, AD
      G(cal, token) {
          return `era_${getLengthName(token.length)}[${cal.year < 0 ? 1 : 0}]`;
      },
      // Year, e.g, 2021; 21
      y(cal, token) {
          return formatYear(cal.year, token);
      },
      // Week year, e.g, 2009; 09
      Y(cal, token) {
          return formatYear(cal.getWeekYear(), token);
      },
      // Month in year, e.g, July; Jul; 07
      M(cal, token) {
          const length = token.length;
          if (length === 1) {
              return '' + cal.month;
          }
          else if (length === 2) {
              return zeroPad(cal.month, 2);
          }
          return `month_${getLengthName(token.length - 2)}[${cal.month - 1}]`;
      },
      // Week in year, e.g, 27
      w(cal, token) {
          return zeroPad(cal.getWeekInYear(), token.length);
      },
      // Week in month, e.g, 2
      W(cal, token) {
          return zeroPad(weekInMonth(cal), token.length);
      },
      // Day in year, e.g, 231
      D(cal, token) {
          return zeroPad(cal.getDayInYear(), token.length);
      },
      // Day in month, e.g, 12
      d(cal, token) {
          return zeroPad(cal.day, token.length);
      },
      // Day of week in month, e.g, 2 (2nd thursday in month)
      F(cal, token) {
          const wim = weekInMonth(cal);
          const dow = cal.getDayOfWeek();
          const firstDowInMonth = new Calendar(new Date(cal.year, cal.month - 1, 1), {
              firstDayOfWeek: cal.firstDayOfWeek
          }).getDayOfWeek();
          return zeroPad(wim - (dow < firstDowInMonth ? 1 : 0), token.length);
      },
      // Day name in week, e.g, Tuesday; Tue
      E(cal, token) {
          return `weekday_${getLengthName(token.length)}[${cal.getDayOfWeek() - 1}]`;
      },
      // Day number of week (1 = Monday, ..., 7 = Sunday)
      u(cal, token) {
          return zeroPad(cal.getDayOfWeek(), token.length);
      },
      // Am/pm marker
      a(cal) {
          return cal.hour < 12 ? 'AM' : 'PM';
      },
      // Hour in day (0-23)
      H(cal, token) {
          return zeroPad(cal.hour, token.length);
      },
      // Hour in day (1-24)
      k(cal, token) {
          return zeroPad(cal.hour === 0 ? 24 : cal.hour, token.length);
      },
      // Hour in am/pm (0-11)
      K(cal, token) {
          return zeroPad(cal.hour % 12, token.length);
      },
      // Hour in am/pm (1-12)
      h(cal, token) {
          const h = cal.hour % 12;
          return zeroPad(h === 0 ? 12 : h, token.length);
      },
      // Minute in hour
      m(cal, token) {
          return zeroPad(cal.minute, token.length);
      },
      // Second in minute
      s(cal, token) {
          return zeroPad(cal.second, token.length);
      },
      // Millisecond
      S(cal, token) {
          return zeroPad(cal.millisecond, token.length);
      },
      // General time zone, e.g, Pacific Standard Time; PST; GMT-08:00
      z(cal, token) {
          const length = token.length;
          const isShort = length < 4;
          const format = isShort ? 'short' : 'long';
          let name = cal.getTimeZoneName({ format });
          if (cal.timeZone === 'UTC' && isShort) {
              const { hours, minutes, sign } = parseTime(cal.offset);
              name += `${sign}${hours}${minutes > 0 ? `:${minutes}` : ''}`;
          }
          return name;
      },
      // RFC 2822 time zone, e.g, -0800
      Z(cal) {
          const offset = cal.offset;
          const { sign, hours, minutes } = parseTime(offset);
          return `${sign}${zeroPad(hours, 2)}${zeroPad(minutes, 2)}`;
      },
      // ISO 8601 time zone, e.g, -08; -0800; -08:00
      X(cal, token) {
          const offset = cal.offset;
          if (offset === 0) {
              return 'Z';
          }
          const { sign, hours, minutes } = parseTime(offset);
          const length = token.length;
          const lead = `${sign}${zeroPad(hours, 2)}`;
          if (length === 1) {
              return lead;
          }
          else if (length === 2) {
              return `${lead}${zeroPad(minutes, 2)}`;
          }
          else if (length === 3) {
              return `${lead}${minutes > 0 ? `:${zeroPad(minutes, 2)}` : ''}`;
          }
          throw new RangeError(`invalid ISO 8601 format: length=${length}`);
      }
  };
  function formatDate(format, date, options) {
      return format.replace(formattingTokensRegExp, (token) => {
          if (token === "''") {
              return "'";
          }
          const firstCharacter = token[0];
          if (firstCharacter === "'") {
              return cleanEscapedString(token);
          }
          const formatter = formatters[firstCharacter];
          if (formatter) {
              return localize(formatter(new Calendar(date, options), token), options);
          }
          return token;
      });
  }
  var index = {
      format(format, date, options = {}) {
          let opts = options;
          let value = Array.isArray(date) ? date[0] : date;
          if (isPlainObject(value) && value.formType === 'field') {
              ({ value, options = {} } = value);
              opts = Object.assign(Object.assign({}, options.dateFormat), opts);
          }
          return formatDate(format, value, Object.assign(Object.assign({}, this.options), opts));
      },
      install(config, options = {}) {
          this.options = options;
          config.plugs.dateFormat = this;
      }
  };

  return index;

}));
